module Screen(
    Packing(..),
    BasicWidget(..),
    doRender,
    tryRender

    ) where

import Control.Exception as E
import Control.Monad.Trans
import Data.Bits
import Control.Concurrent
import Data.IORef
import Monad

import ErrorLog
import Curses
import Format
import GenUtil



data Canvas = Canvas {window :: Window, origin :: (Int,Int), bounds :: (Int,Int)  }
data Widget = Widget {
    render :: Canvas -> IO (), 
    getBounds :: IO (Int,Int), 
    hasChanged :: IO Bool, 
    changedSignal :: Signal (),
    processKey :: Key -> IO Bool
    }
data AnyWidget = forall w . BasicWidget w => AnyWidget w
newtype Alternate = Alternate (IORef (Widget,Bool))
-- data Switch = Switch (MVar AnyWidget)
data Packing = NoExpand | Expand | ExpandFill
data Attribute = AttrBold | AttrBlink | AttrDim | AttrReverse | AttrUnderline | AttrColor String String

instance Show Canvas where
    show c = fmtSs "<canvas:origin=%s,bounds=%s>" [show (origin c), show (bounds c)]


class BasicWidget w where
    widgetEmpty :: w
    widgetEmptySized :: Float -> Float -> w
    widgetCenter :: w -> w
    widgetSimpleFrame :: w -> w
    widgetVerticalBox :: Bool -> [(Packing, w)] -> w
    widgetHorizontalBox :: Bool -> [(Packing, w)] -> w
    widgetText :: String -> w
    widgetAttr :: [Attribute] -> w -> w
    widgetScroll :: w -> IO w
    widgetNewSwitch :: IO (w,MVar w)
    widgetProcessKey :: w -> Key -> IO Bool


instance BasicWidget Widget where
    widgetEmpty = emptyWidget
    widgetEmptySized x y = emptySizedWidget (floor (x + 0.5)) (floor (y + 0.5))
    widgetCenter w = centerWidget w 
    widgetSimpleFrame w = boxWidget w
    widgetVerticalBox True ws = homVerticalBox ws
    widgetVerticalBox False ws = verticalBox ws []
    widgetHorizontalBox True ws = homHorizontalBox ws
    widgetHorizontalBox False ws = horizontalBox ws []
    widgetText s = staticText s
    widgetAttr al w = attrWidget (foldl (flip ($)) attr0 (map ma al)) w where
	ma AttrBlink a = a `setBlink` True
	ma AttrBold a = a  `setBold` True
	ma AttrDim a = a  `setDim` True
	ma AttrReverse a = a  `setReverse` True
	ma AttrUnderline a = a `setUnderline` True
	ma _ a = a
    widgetProcessKey w = processKey w


instance BasicWidget () where
    widgetEmpty = ()
    widgetEmptySized x y = ()
    widgetCenter w = ()
    widgetSimpleFrame w = ()
    widgetVerticalBox _ ws = ()
    widgetHorizontalBox _ ws = ()
    widgetText s = ()
    widgetAttr _ _ = ()



isExpand NoExpand = False
isExpand _ = True

isFill ExpandFill = True
isFill _ = False



doRender w = do
    (ys,xs) <- scrSize
    erase 
    let c = (Canvas {window = stdScr, origin = (0,0), bounds = (xs,ys)})  
    eannM ("doRender " ++ show c) $ render w c
    refresh

tryRender w = do
    b <- hasChanged w
    when b $ doRender w
    



-------------------
-- Drawing Routines
-------------------

canvasTranslate :: Canvas -> (Int,Int) -> Canvas
canvasTranslate canvas@(Canvas {origin = (x,y), bounds = (xb,yb)}) (xt,yt) = assert (xt > 0 && yt > 0) $ canvas {origin = (x + xt, y + yt), bounds = (max (xb - xt) 0, max (yb - yt) 0)}


withAttr :: Canvas -> Attr -> IO a -> IO a
withAttr canvas attr action = E.bracket (wAttrGet w) (wAttrSet w) f where
    f (a,p) = wAttrSet w (a .|. attr,p) >> action
    w = window canvas


drawString :: Canvas -> String -> IO ()
drawString canvas s = eannM (fmtSs "drawString %s %s" [(show canvas), s]) $
    when (yb > 0) $ attempt $ mvWAddStr (window canvas) (snd $ origin canvas) (fst $ origin canvas) (take xb s) where
	(xb,yb) = bounds canvas


boundCanvas :: Maybe Int -> Maybe Int -> Canvas -> Canvas 
boundCanvas mx my canvas = (canvas {bounds = (nx,ny)}) where
    (cx,cy) = bounds canvas
    nx = maybe cx (min cx) mx
    ny = maybe cy (min cy) my


-- rendering routines

renderCentered :: Widget -> Canvas -> IO ()
renderCentered w canvas = do
	(xs,ys) <- getBounds w
	let (xs',ys') =  bounds canvas
	    f cs ws | ws >= cs = 0
	    f cs ws = (cs - ws) `div` 2
	    nc = (canvasTranslate canvas (f xs' xs,f ys' ys))
	renderChild w nc {bounds=(min xs $ fst $ bounds nc ,min ys $ snd $ bounds nc)} 

renderBounded :: Widget -> Canvas -> Maybe Int -> Maybe Int -> IO ()
renderBounded w canvas mx my = renderChild w (canvas {bounds = (nx,ny)}) where
    (cx,cy) = bounds canvas
    nx = maybe cx (min cx) mx
    ny = maybe cy (min cy) my

renderChild :: Widget -> Canvas -> IO ()
renderChild _ (Canvas {bounds = (xb,yb)}) | xb == 0 || yb == 0 = return ()
renderChild w c = render w c

----------
-- Widgets
----------

emptyWidget = Widget {
    render = const (return ()), 
    getBounds = return (0,0), 
    hasChanged = return False, 
    changedSignal = nonSignal,
    processKey = \_ -> return False
    }
emptySizedWidget x y =  emptyWidget {getBounds = return (x,y)}


-- layout widgets

-- | centers child widget, causes child to not use any extra space.
centerWidget :: Widget -> Widget
centerWidget w = w {render = rst} where
    rst canvas = do 
	(xs,ys) <- getBounds w
	let (xs',ys') =  bounds canvas
	renderChild w (canvasTranslate canvas (f xs' xs,f ys' ys)) {bounds=(min xs $ fst $ bounds canvas ,min ys $ snd $ bounds canvas)} where
	    f cs ws | ws >= cs = 0
	    f cs ws = (cs - ws) `div` 2


-- General Widgets

-- | draw box around child widget
boxWidget :: Widget -> Widget
boxWidget w = w {render = rst, getBounds = gb} where
    gb = getBounds w >>= \(x,y) -> return (x + 2, y + 2)
    rst canvas@(Canvas {bounds = (xb,yb)}) = do
	let tb = '+' : (replicate (xb - 2) '-' ++ "+")
	    bc c@(Canvas {bounds = (xb,yb)}) = c {bounds = (max (xb - 1) 0, max (yb - 1) 0)}
	renderChild w (canvasTranslate (bc canvas) (1,1))	
	drawString canvas tb
	drawString (canvasTranslate canvas (0,yb - 1)) tb
	sequence_ [drawString (canvasTranslate canvas (xt,yt)) "|"| yt <- [1..yb - 2], xt <- [0,xb - 1]]


{-
widgetMinBBox :: Maybe Int -> Maybe Int -> Widget -> Widget
widgetMaxBBox :: Maybe Int -> Maybe Int -> Widget -> Widget
widgetExactBBox :: Maybe Int -> Maybe Int -> Widget -> Widget
-}
    
staticText s = emptyWidget {render = rst, getBounds = return (length s,1)  } where
    rst canvas = drawString canvas s 


filledSizedWidget :: Int -> Int -> Char -> Widget
filledSizedWidget 0 y _ = (emptySizedWidget 0 y) 
filledSizedWidget x 0 _ = (emptySizedWidget x 0) 
filledSizedWidget x y c = (emptySizedWidget x y) {render = rnd} where
    rnd canvas = eannM (fmt "filledSizedWidget %i %i %/" [fi x, fi y, fa c]) $ ds canvas (min yb y)
     where
	ds _ 0 = return ()
	ds canvas yb = do 
	    drawString canvas (replicate (min xb x) c) 
	    ds (canvasTranslate canvas (0,1)) (yb - 1) 
	(xb,yb) = bounds canvas
	
cMin (x,y) = liftT2 (min x,min y)
cMax (x,y) = liftT2 (max x,max y)



verticalBox :: [(Packing,Widget)] -> [(Packing,Widget)] -> Widget
verticalBox starts ends = emptyWidget {render = rd, getBounds = gb, hasChanged = hc} where
    awpb = starts ++ ends
    awp = if any isExpand $ fsts awpb then awpb else starts ++ [(Expand,emptyWidget)] ++ ends
    aw = snds awp
    hc = fmap or (mapM hasChanged aw)
    gb = do
	(xs,ys) <- fmap unzip $ mapM getBounds aw
	return (maximum xs, sum ys)
    rd canvas = do
	let (xb,yb) = bounds canvas
	(xs,ys) <- gb
	let ne = length (filter isExpand $ fsts awp)
	    el = if yb > ys then splitSpace (yb - ys) ne else replicate ne 0
	    f ((e:es),canvas) (ps,w) | isExpand ps = do
		(wbx,wby) <- getBounds w
		if isFill ps then 
		    renderChild w (boundCanvas Nothing (Just (wby + e)) canvas)
			else renderCentered w (boundCanvas Nothing (Just (wby + e)) canvas)
		return (es,canvasTranslate canvas (0,wby + e))
	    f (es,canvas) (ps,w) = do
		(wbx,wby) <- getBounds w
		renderCentered w (boundCanvas Nothing (Just wby) canvas)
		return (es,canvasTranslate canvas (0,wby))
	foldlM_ f (el,canvas) awp


horizontalBox :: [(Packing,Widget)] -> [(Packing,Widget)] -> Widget
horizontalBox starts ends = emptyWidget {render = rd, getBounds = gb, hasChanged = hc} where
    awpb = starts ++ ends
    awp = if any isExpand $ fsts awpb then awpb else starts ++ [(Expand,emptyWidget)] ++ ends
    aw = snds awp
    hc = fmap or (mapM hasChanged aw)
    gb = do
	(xs,ys) <- fmap unzip $ mapM getBounds aw
	return (sum xs, maximum ys)
    rd canvas = do
	let (xb,yb) = bounds canvas
	(xs,ys) <- gb
	let ne = length (filter isExpand $ fsts awp)
	    el = if xb > xs then splitSpace (xb - xs) ne else replicate ne 0
	    f ((e:es),canvas) (ps,w) | isExpand ps = do
		(wbx,wby) <- getBounds w
		if isFill ps then 
		    renderChild w (boundCanvas (Just (wbx + e)) Nothing canvas)
			else renderCentered w (boundCanvas (Just (wbx + e)) Nothing canvas)
		return (es,canvasTranslate canvas (wbx + e,0))
	    f (es,canvas) (ps,w) = do
		(wbx,wby) <- getBounds w
		renderCentered w (boundCanvas (Just wbx) Nothing canvas)
		return (es,canvasTranslate canvas (wbx,0))
	foldlM_ f (el,canvas) awp

homVerticalBox :: [(Packing,Widget)] -> Widget
homVerticalBox awp = emptyWidget {render = rd, getBounds = gb, hasChanged = hc} where
    hc = fmap or (mapM hasChanged aw)
    aw = snds awp
    gb = do
	(xs,ys) <- fmap unzip $ mapM getBounds aw
	return (maximum xs, (maximum ys) * length aw)
    rd canvas = do
	let (xb,yb) = bounds canvas
	    el = splitSpace yb (length aw) 
	    f ((e:es),canvas) (ps,w)  = do
		if isFill ps then 
		    renderChild w (boundCanvas Nothing (Just e) canvas)
			else renderCentered w (boundCanvas Nothing (Just e) canvas)
		return (es,canvasTranslate canvas (0,e))
	foldlM_ f (el,canvas) awp


homHorizontalBox :: [(Packing,Widget)] -> Widget
homHorizontalBox awp = emptyWidget {render = rd, getBounds = gb, hasChanged = hc} where
    hc = fmap or (mapM hasChanged aw)
    aw = snds awp
    gb = do
	(xs,ys) <- fmap unzip $ mapM getBounds aw
	return (maximum xs * length aw, maximum ys)
    rd canvas = do
	let (xb,yb) = bounds canvas
	    el = splitSpace xb (length aw) 
	    f ((e:es),canvas) (ps,w)  = do
		if isFill ps then 
		    renderChild w (boundCanvas (Just e) Nothing canvas)
			else renderCentered w (boundCanvas (Just e) Nothing canvas)
		return (es,canvasTranslate canvas (e,0))
	foldlM_ f (el,canvas) awp

splitSpace m n = f (m `mod` n) $ replicate n (m `div` n) where
    f 0 xs = xs
    f n (x:xs) = (x + 1) : f (n - 1) xs
    f n' [] = errorf "splitSpace %i %i (f %i []): this can't happen." [fi m, fi n, fi n']
	



	

newAlternate :: IO (Widget, Alternate)
newAlternate = do
    r <- newIORef (emptyWidget,False)
    let w = emptyWidget {render = rw, getBounds = gb, hasChanged = hc} 
	rw canvas = do
	    (w,_) <- readIORef r
	    renderChild w canvas
	    writeIORef r (w,False)
	gb = do
	    (w,_) <- readIORef r
	    getBounds w 
	hc = do
	    (w,c) <- readIORef r
	    if c then return True else hasChanged w
    return (w,Alternate r)
    

switchAlternate :: Alternate -> Widget -> IO ()
switchAlternate (Alternate r) w = writeIORef r (w,True)


attrWidget :: Attr -> Widget -> Widget
attrWidget a w = w {render = nr} where
    nr canvas = withAttr canvas a (renderChild w  canvas)

--------------------
-- Signals and SVars
--------------------

data Signal a = Signal {
    connect :: MonadIO m =>  (a -> IO ()) -> m (),
    disconnectAll :: MonadIO m => m (),
    signal :: MonadIO m => a -> m ()
    }


-- signals

newSignal :: MonadIO m => m (Signal a)
newSignal = do
    mv <- liftIO $ newMVar ([]::[a -> IO ()])
    let c action = liftIO (modifyMVar_ mv (\as -> return (action:as))) 
	d :: MonadIO m => m ()
	d = (liftIO $ swapMVar mv []) >> return ()
	s v =  do
	    cl <- liftIO $ readMVar mv
	    liftIO $ mapM_ ($ v) cl
    return $ Signal {connect = c, disconnectAll = d, signal = s} 


    
nonSignal = Signal {
    connect = \_ -> return (),
    disconnectAll = return (),
    signal = \_ -> return ()
    }

signalCondenser :: MonadIO m => [Signal a] -> m (Signal a)
signalCondenser sl = do
    s <- newSignal
    mapM_ (\s' -> connect s' (signal s)) sl
    return s


-- SVars

data SVar a = SVar (Signal a) (MVar a)


svarSignal :: SVar a -> Signal a
svarSignal (SVar s _) = s

newSVar :: MonadIO m => a -> m (SVar a)
newSVar v = do
    s <- newSignal 
    mv <- liftIO $ newMVar v
    return $ SVar s mv
writeSVar :: MonadIO m => (SVar a) -> a -> m ()
writeSVar (SVar s mv) v = do
    liftIO $ modifyMVar_ mv (const (return v))
    signal s v
modifySVar :: MonadIO m => (SVar a) -> (a -> a) -> m ()
modifySVar (SVar s mv) f = do
    v <- liftIO $ modifyMVar mv (\x -> (let y = f x in return (y,y)))
    signal s v
readSVar :: MonadIO m => (SVar a) -> m a
readSVar (SVar _ mv) = liftIO $ readMVar mv  

    



{- Attic

newtype KeyHandler = KeyHandler (Key -> IO KeyHandler)
unKeyHandler (KeyHandler kh) = kh

{-# NOINLINE loopExit_r #-}
loopExit_r = unsafePerformIO $ newMVar False

exitLoop :: IO ()
exitLoop = swapMVar loopExit_r True >> return ()

runLoop :: KeyHandler -> Widget -> IO ()
runLoop kh w = gProgram $ \canvas ->  rp canvas >> loop canvas kh where
	loop canvas kh = do
	    hc <- hasChanged w
	    when hc $ rp canvas
	    k <- getCh
	    kh <- unKeyHandler kh k
	    c <- readMVar loopExit_r 
	    unless c $ loop canvas kh
	gProgram :: (Canvas -> IO a) -> IO a 
	gProgram action = E.bracket_ initCurses endWin $ do
	   ss <- scrSize 
	   action $ Canvas {window = stdScr, origin = (0,0), bounds = ss}
	rp canvas = erase >> render w canvas >> refresh

------------
-- Test Code
------------

nkh a = KeyHandler f where
    f (KeyChar 'q') = exitLoop >> (return $ nkh a)
    f key@(KeyF i) = do
	switchAlternate a (simpleVerticalBox $ replicate i (staticText (show key)))
	(return $ nkh a)
    f key = do
	switchAlternate a (staticText (show key))
	return $ nkh a
    
	

screenMain = do
    mainThd <- myThreadId
    Posix.installHandler Posix.sigINT (Posix.Catch (killThread mainThd)) Nothing
    (w,a) <- newAlternate
    switchAlternate a (staticText "Sup?")
    runLoop (nkh a) (simpleVerticalBox [w,staticText "Is the keystroke."])

-------------
-- Disposable
-------------


class Disposable a where 
    dispose :: a -> IO ()
    
withDisposable :: (MonadIO m, Disposable a) => m a -> (a -> m b) -> m b
withDisposable c a = do
    v <- c
    r <- a v
    liftIO $ dispose v
    return r
    

-}
